/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */


#ifndef _COMPILER_H_
#define _COMPILER_H_


#include "x86tsc.h"

#include "opcodes.h"
#define opc_invokeignored_nocheck	230	/* 0xe6 */

#define opc_getstatic2	231	/* 0xe7 */
#define opc_putstatic2	232
#define opc_getfield2	233
#define opc_putfield2	234

#define opc_methodhead	235	/* 0xeb */
#define opc_exception_handler	236
#define opc_methodtail	237

#define opc_fill_cache	238	/* 0xee */
#define opc_array_check	239

#define opc_iastore1	240	/* 0xf0 */
#define opc_lastore1	241

#define opc_stateto0	242	/* 0xf2 */
#define opc_stateto1	243
#define opc_stateto2	244
#define opc_stateto3	245
#define opc_stateto4	246

#define opc_goto_st0	247	/* 0xf7 */
#define opc_goto_st1	248
#define opc_goto_st2	249
#define opc_goto_st3	250
#define opc_goto_st4	251

#define NOPCODES	256


/* offerred by Sun */
#include "native.h"


/* Additional type definition */

#if !(defined(_ILP32) || defined(_LP64))
#  ifndef _UINT16_T
#    define _UINT16_T
typedef unsigned short uint16_t;
#    endif	/* _UINT16_T */

#  ifndef __BIT_TYPES_DEFINED__
#    ifndef _INT16_T
#      define _INT16_T
typedef short int16_t;
#    endif	/* _INT16_T */
#  endif	/* __BIT_TYPES_DEFINED__ */

#endif	/* _ILP32, LP64 */

#  ifdef __FreeBSD__
typedef u_int16_t	uint16_t;
#  endif	/* __FreeBSD__ */


/* Additional macro definition */
#define handleToClassClass(h)	(java_lang_Object_getClass((HObject *)(h)))


/*
 * features
 */

#define EXC_BY_SIGNAL
#define GET_SIGCONTEXT
#undef IGNORE_DISABLE

#ifdef RUNTIME_DEBUG
#undef EXC_BY_SIGNAL
#endif

/*
 * OS dependent macro definition
 */

#if defined (linux)
#  define EXECUTEJAVA_IN_ASM
#  if defined __GLIBC__ && __GLIBC__ >= 2
#    undef  RESOLVE_SYMBOL_ON_CODE
#  else
#    define RESOLVE_SYMBOL_ON_CODE
#  endif
#  define SEARCH_SIGCONTEXT
#elif defined (__FreeBSD__)
#  undef  EXECUTEJAVA_IN_ASM
#  undef  RESOLVE_SYMBOL_ON_CODE
#  undef  SEARCH_SIGCONTEXT
#else
#  undef  EXECUTEJAVA_IN_ASM
#  undef  RESOLVE_SYMBOL_ON_CODE
#  define  SEARCH_SIGCONTEXT
#endif


#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
#  if defined(linux)
#    include <asm/sigcontext.h>	/* for struct sigcontext */
#    include <linux/version.h>	/* for kernel version */
#    if LINUX_VERSION_CODE < ((2<<16)+(1<<8)+(0))
typedef struct sigcontext_struct sigcontext;	/* 2.0.X */
#    else
typedef struct sigcontext sigcontext;		/* 2.1.X, 2.2.X */
#    endif
#    define SC_ECX	ecx
#    define SC_EDX	edx
#    define SC_ESI	esi
#    define SC_EIP	eip
#    define SC_EBP	ebp
#    define SC_ESP	esp
#  elif defined(__FreeBSD__)
#    include <machine/signal.h>	/* for struct sigcontext */
typedef struct sigcontext sigcontext;
#    define SC_ECX	sc_ecx
#    define SC_EDX	sc_edx
#    define SC_ESI	sc_esi
#    define SC_EIP	sc_eip
#    define SC_EBP	sc_ebp
#    define SC_ESP	sc_esp
#  else
typedef struct sigcontext sigcontext;
#    define SC_ECX	ecx
#    define SC_EDX	edx
#    define SC_ESI	esi
#    define SC_EIP	eip
#    define SC_EBP	ebp
#    define SC_ESP	esp
#  endif
#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */


#ifdef CODE_DB
#  undef CODE_DB_DEBUG

#  define CODEDB_PREFIX		"shujit-code"
#  define CODEDB_DB_SUFFIX	".db"
#  define CODEDB_PAGE_SUFFIX	".page"

#  define CODEDB_PAGE	CODEDB_PREFIX CODEDB_PAGE_SUFFIX
#  ifdef GDBM
#    define LIBDBM	"libgdbm.so"
#    define CODEDB_DB	CODEDB_PREFIX CODEDB_DB_SUFFIX
#  else
#    define LIBDBM	"libndbm.so"
#    define CODEDB_DB	CODEDB_PREFIX
#  endif
#endif	/* CODE_DB */


/*
 * Select appropriate malloc library
 */
# define sysMalloc	malloc
# define sysFree	free
# define sysCalloc	calloc
# define sysRealloc	realloc


/*
 * Macros
 */

#define CREDIT "  shuJIT  for Sun JVM/IA-32  Copyright 1998,1999 by SHUDO Kazuyuki\n"

#define COMPILER_VERSION	5

#define CODESIZE_FNAME	"jit_codesize"

/* STR(macro) is permuted to "value of the macro" */
#define _STR(T)	#T
#define STR(MACRO)	_STR(MACRO)


/*
 * Type definition
 */

/* CatchFrame_w_state is based on CatchFrame */
/* Requirement for C compiler:
	sizeof(CatchFrame_w_state) equals sizeof(CatchFrame) */
typedef struct CatchFrame_w_state {
    long    start_pc, end_pc;
    long    handler_pc;
    void*   compiled_CatchFrame;
    short   catchType;
    short   state;	/* added by SHUDO */
} CatchFrame_w_state;


#ifdef EXC_BY_SIGNAL
typedef struct throw_entry {
  uint32_t start;
  uint16_t len;
  uint16_t byteoff;
} throwentry;
#endif	/* EXC_BY_SIGNAL */

typedef struct compiled_code_info {
  int32_t code_size;
  int32_t ret_size;
  char *argsizes;

  uint32_t exc_handler_nativeoff;
  uint32_t finish_return_nativeoff;

#ifdef EXC_BY_SIGNAL
#define DEFAULT_THROWTABLE_SIZE	1024
  throwentry *throwtable;
  uint32_t throwtablelen;
  uint32_t throwtablesize;
#endif	/* EXC_BY_SIGNAL */
} CodeInfo;


typedef struct pcentry {
  uint16_t opcode;
  uint16_t state;
  uint32_t byteoff;
  uint32_t nativeoff;
} pcentry;

typedef struct jpentry {
  unsigned int tgtoff;
  unsigned int argoff;
} jpentry;

typedef struct compiler_context {
  ExecEnv *ee;
  struct methodblock *mb;

  /* buffer for compiled code */
#define DEFAULT_BUF_SIZE	4096
  unsigned char *buffer;
  int buf_size;
  unsigned char *bufp;

  /* program counter table */
#define DEFAULT_PCTABLE_SIZE	128
  int pctablesize;
  uint32_t pctablelen;
  pcentry *pctable;

  /* jump instruction table */
#define DEFAULT_JPTABLE_SIZE	128
  int jptablesize;
  int jptablelen;
  jpentry *jptable;
} CompilerContext;


/*
 * Global Variables
 */
/* in compiler.c */

#ifndef IGNORE_DISABLE
extern bool_t compiler_enabled;
#endif	/* IGNORE_DISABLE */

#if (defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)) && defined(SEARCH_SIGCONTEXT)
#  ifdef linux
extern unsigned short reg_gs;
extern unsigned short reg_fs;
#  endif
extern unsigned short reg_es;
extern unsigned short reg_ds;
#  ifdef __FreeBSD__
extern unsigned short reg_cs;
extern unsigned short reg_ss;
#  endif
#endif	/* (EXC_BY_SIGNAL || GET_SIGCONTEXT) && SEARCH_SIGCONTEXT */

#ifdef CODE_DB
#  ifdef GDBM
#    include <gdbm.h>
extern GDBM_FILE db;
#  else
#    include <ndbm.h>
extern DBM *db;
#  endif
extern int db_page;
#endif	/* CODE_DB */

extern int opt_quiet;
	/* suppress initial message and some outputs */
extern int opt_outcode;
	/* write generated code to code_<classname>_<methodname>.s */
extern int opt_codesize;
	/* write code size of each methods to the file jit_codesize */
extern int opt_dontcmplvmcls;
	/* suppress compilation classes
	   which is already loaded when JIT is initialized */
extern int opt_igndisable;
	/* make java.lang.Compiler#disable() void */
extern int opt_cmplclinit;
	/* compile class initializer */
#ifdef CODE_DB
extern int opt_codedb;
	/* save and re-use generated native code */
#endif	/* CODE_DB */
extern int opt_cmplatload;
	/* compile the whole class when the class is loaded */

extern void *sym_compileAndInvokeMethod;
extern void *sym_invokeJITCompiledMethod;
extern void *sym_invokeJavaMethod;
extern void *sym_invokeSynchronizedJavaMethod;
extern void *sym_invokeAbstractMethod;
extern void *sym_invokeLazyNativeMethod;
#ifdef CODE_DB
#  ifdef GDBM
extern GDBM_FILE (*sym_dbm_open)(char *,int,int,int,void (*)());
extern void (*sym_dbm_close)(GDBM_FILE);
extern int (*sym_dbm_store)(GDBM_FILE,datum,datum,int);
extern datum (*sym_dbm_fetch)(GDBM_FILE,datum);
extern void (*sym_dbm_sync)(GDBM_FILE);
#  else
extern DBM *(*sym_dbm_open)(const char *,int,int);
extern void (*sym_dbm_close)(DBM *);
extern int (*sym_dbm_store)(DBM *,datum,datum,int);
extern datum (*sym_dbm_fetch)(DBM *,datum);
#  endif
#endif	/* CODE_DB */

#ifdef JITDO
/* in proxy.c */
extern struct methodtable *proxy_methodtable;	/* in proxy.c */
#endif	/* JITDO */


/*
 * Global Functions
 */
/* in code.c */
extern volatile void assembledCode(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee
#ifdef RUNTIME_DEBUG
	,int runtime_debug
#endif
);

/* in computil.c */
extern void *access2invoker(int access);

extern void showCompilerContext(CompilerContext *cc);
extern CompilerContext *newCompilerContext(struct methodblock *mb);
extern void freeCompilerContext(CompilerContext *cc);
extern void resetCompilerContext(CompilerContext *cc);

extern void writeToBuffer(CompilerContext *cc, void *, size_t);
extern void cancelOnBuffer(CompilerContext *cc, size_t);
extern void ensureBufferSize(CompilerContext *cc, size_t req);

#ifdef CODE_DB
extern void pctableExtend(CompilerContext *cc, uint32_t size);
#endif	/* CODE_DB */
extern void pctableAdd(CompilerContext *cc,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff);
extern void pctableInsert(CompilerContext *cc, int index,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff);
extern void pctableDelete(CompilerContext *cc, int index);
extern pcentry *pctableGet(CompilerContext *cc, int byteoff);

extern void jptableAdd(CompilerContext *cc,
	unsigned int tgtoff, unsigned int argoff);

extern CodeInfo *prepareCompiledCodeInfo(ExecEnv *ee, struct methodblock *mb);
extern void freeCompiledCodeInfo(CodeInfo *info);

#ifdef EXC_BY_SIGNAL
#ifdef CODE_DB
extern void throwtableExtend(CodeInfo *info, uint32_t size);
#endif	/* CODE_DB */
extern void throwtableAdd(CodeInfo *info,
	uint32_t start, uint16_t len, uint16_t byteoff);
extern throwentry *throwtableGet(CodeInfo *info, uint32_t nativeoff);
#endif	/* EXC_BY_SIGNAL */


/* in invoker.c */
extern bool_t compileAndInvokeMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee);
extern bool_t invokeJITCompiledMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee);

/* in compile.c */
extern int compileMethod(struct methodblock *mb);
extern void freeMethod(struct methodblock *mb);
#ifdef RESOLVE_SYMBOL_ON_CODE
extern void initFunctionSymbols();
#endif

#ifdef CODE_DB
/* in codedb.c */
void writeCompiledCode(
#  ifdef GDBM
GDBM_FILE db
#  else
DBM *db
#  endif
, int fd, CompilerContext *cc);
int readCompiledCode(
#  ifdef GDBM
GDBM_FILE db
#  else
DBM *db
#  endif
, int fd, CompilerContext *cc);
#endif	/* CODE_DB */

/* in runtime.c */
#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
extern void showSigcontext(sigcontext *sc);
#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */
extern int invokeMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	JHandle *obj, struct methodblock *method, int args_size, ExecEnv *ee,
	int retsize, int bytepcoff, stack_item *stackpointer
);
extern struct methodblock *getInterfaceMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee,
	JHandle *obj, struct methodblock *imethod, unsigned char *guessptr);
extern JHandle *multianewarray(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee, int dimensions, ClassClass *arrayclazz,
	stack_item *stackpointer);
extern struct CatchFrame *searchCatchFrame(ExecEnv *ee, struct methodblock *mb,
	int bytepcoff
#ifdef RUNTIME_DEBUG
	, int runtime_debug
#endif
);
extern void showStackFrames(ExecEnv *ee);
#ifdef RUNTIME_DEBUG
extern void showExcStackTrace(JHandle *throwable);
extern char *showObjectBody(char *sig, JHandle *obj);
#endif

#ifdef JITDO
/* in proxy.c */
#include "jitdo/NET_shudo_jitdo_VMAddress_old.h"
extern JHandle *proxy_new(ExecEnv *ee,
	HNET_shudo_jitdo_VMAddress *addr, ClassClass *cb);

extern int32_t proxy_get32field(ExecEnv *, JHandle *proxy, int32_t slot);
extern int64_t proxy_get64field(ExecEnv *, JHandle *proxy, int32_t slot);
extern JHandle *proxy_getobjfield(ExecEnv *, JHandle *proxy, int32_t slot);
extern void proxy_set32field(ExecEnv *, JHandle *p, int32_t slot, int32_t v);
extern void proxy_set64field(ExecEnv *, JHandle *p, int32_t slot, int64_t v);
extern void proxy_setobjfield(ExecEnv *, JHandle *p, int32_t slot, JHandle *v);

extern int proxy_invoke(ExecEnv *ee, JHandle *proxy,
	char *name, char *sig, stack_item *stackpointer);

/* in copyable.c */
extern bool_t isCopyable(ExecEnv *ee, ClassClass *clazz);
#endif	/* JITDO */

#endif /* _COMPILER_H_ */
