/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include "compiler.h"
#include "monitor.h"	/* for monitor{Enter,Exit}() and macro BINCLASS_*() */

#ifdef JITDO
#include "jitdo/NET_shudo_jitdo_Proxy_old.h"
#endif	/* JITDO */


#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
/*
 * show contents of structure sigcontext
 */
void showSigcontext(sigcontext *sc) {
  printf("EAX: %08x, ECX: %08x, EDX: %08x, EBX: %08x\n",
#if defined(linux)
	sc->eax, sc->ecx, sc->edx, sc->ebx
#elif defined(__FreeBSD__)
	sc->sc_eax, sc->sc_ecx, sc->sc_edx, sc->sc_ebx
#endif
  );

  printf("ESI: %08x, EDI: %08x\n",
#if defined(linux)
	sc->esi, sc->edi
#elif defined(__FreeBSD__)
	sc->sc_esi, sc->sc_edi
#endif
  );

  printf("ESP: %08x, EBP: %08x EIP: %08x\n",
#if defined(linux)
	sc->esp, sc->ebp, sc->eip
#elif defined(__FreeBSD__)
	sc->sc_esp, sc->sc_ebp, sc->sc_eip
#endif
  );

#if defined(linux)
  printf("ESP at signal: %08x\n", sc->esp_at_signal);
#endif

#if defined(linux)
  printf("SS: %04x, CS: %04x, DS: %04x, ES: %04x, FS: %04x, GS: %04x\n",
	sc->ss, sc->cs, sc->ds, sc->es, sc->fs, sc->gs
#elif defined(__FreeBSD__)
  printf("SS: %04x, CS: %04x, DS: %04x, ES: %04x\n",
	sc->sc_ss & 0xffff, sc->sc_cs & 0xffff,
	sc->sc_ds & 0xffff, sc->sc_es & 0xffff
#endif
  );

  fflush(stdout);
}
#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */


/*
 * returns: return value size. -1 if an exception occurred.
 */
int invokeMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	JHandle *obj, struct methodblock *method, int args_size, ExecEnv *ee,
	int retsize, int bytepcoff, stack_item *stackpointer
) {
  JavaFrame *cur_frame;	/* ee->current_frame */
  int access;
#ifdef RUNTIME_DEBUG
  CodeInfo *info;
  info= (CodeInfo *)method->CompiledCodeInfo;
#endif

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("  native TOS: 0x%x\n", (int)stackpointer);
    fflush(stdout);
  }
#endif

  /* set ee->current_frame->lastpc */
  cur_frame = ee->current_frame;
  cur_frame->lastpc = cur_frame->current_method->code + bytepcoff;

  access = method->fb.access;
  stackpointer += (args_size - 1);

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    struct methodblock *cur_mb = cur_frame->current_method;

    printf("  ee: 0x%08x\n", (int)ee);
    printf("  current_frame: 0x%08x\n", (int)cur_frame);
    printf("  args_size, retsize, nlocals: %d, %d, %d\n",
		args_size, retsize, method->nlocals);
    printf("  obj: 0x%08x\n", (int)obj);
    printf("  method: 0x%08x\n", (int)method);
    printf("  caller method:");
    fflush(stdout);
    if (cur_mb)
      printf(" %s#%s %s",
		cbName(cur_mb->fb.clazz),
		cur_mb->fb.name, cur_mb->fb.signature);
    else
      printf(" (null)");
    fflush(stdout);
    printf(" (0x%08x)\n", (int)cur_mb);
    printf("  bytepcoff: 0x%x(%d)\n", bytepcoff, bytepcoff);
    fflush(stdout);
  }
#endif

#ifdef RUNTIME_DEBUG
  /* let user know calling normal Java method */
  if ((void *)method->invoker == sym_invokeJavaMethod) {
    printf("call normal Java method: %s#%s\n",
		cbName(method->fb.clazz), method->fb.name);
    fflush(stdout);
  }
  else if ((void *)method->invoker == sym_invokeSynchronizedJavaMethod) {
    printf("call normal sync. Java method: %s#%s\n",
		cbName(method->fb.clazz), method->fb.name);
    fflush(stdout);
  }
#endif	/* RUNTIME_DEBUG */


#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    char *sig = method->fb.signature;
    char *name = method->fb.name;
    stack_item *optop = stackpointer;
    int i;

    printf("  obj: ");
    if (access & ACC_STATIC)
      printf("%s", unhand((Hjava_lang_Class *)obj)->name);
    else if (obj_flags(obj) == T_NORMAL_OBJECT)
      printf("%s", cbName(obj->methods->classdescriptor));
    else	/* array object */
      printf("array");
    printf(" (0x%08x)\n", (int)obj);
    printf("  method: %s#%s %s\n", cbName(fieldclass(&method->fb)), name, sig);
    printf("  ret val size: %d\n", info->ret_size);
    printf("  acc: 0x%x", access);
    if (access & ACC_NATIVE)  printf(" native");
    if (access & ACC_MACHINE_COMPILED)  printf(" machine_compiled");
    if (access & ACC_PUBLIC)  printf(" public");
    if (access & ACC_PRIVATE)  printf(" private");
    if (access & ACC_PROTECTED)  printf(" protected");
    if (access & ACC_STATIC)  printf(" static");
    if (access & ACC_FINAL)  printf(" final");
    if (access & ACC_SYNCHRONIZED)  printf(" synchronized");
    if (access & ACC_ABSTRACT)  printf(" abstract");
    printf("\n");
    printf("  args_size: %d\n", args_size);
    fflush(stdout);
    i = 0;
    if (!(access & ACC_STATIC)) {
      printf("    0x%08x @ 0x%08x  ", (int)optop[0].h, (int)optop);
      fflush(stdout);
      showObjectBody("L;", optop[0].h);
      printf("\n");
      i++;
    }
    if (sig[0] == '(')  sig++;
    for (; i < args_size; i++) {
      JHandle *arg = optop[-i].h;
      if ((sig[0] == 'D') || (sig[0] == 'J'))  i++;
      printf("    0x%08x @ 0x%08x  ", (int)arg, (int)(optop - i));
      fflush(stdout);
      sig = showObjectBody(sig, arg);
      printf("\n");
    }
    fflush(stdout);
  }
#endif	/* RUNTIME_DEBUG */


#if 0
#define CALL_INVOKER \
    __asm__("pushl 24(%ebp)");\
    __asm__("movl  %0,%%eax" : : "m" (method->invoker));\
    __asm__("pushl 20(%ebp)\n\tpushl 16(%ebp)\n\tpushl 12(%ebp)");\
    __asm__("call  *%eax\n\t"\
	    "addl  $16,%esp");\
  {\
    register int ret asm("eax");\
    if (!ret)  return -1;\
  }
#else
#define CALL_INVOKER \
  if (!(method->invoker(obj, method, args_size, ee)))\
    return -1
	/* obj is required for enter/exit a monitor of the object. */
#endif

  if ((access & ACC_MACHINE_COMPILED)
	/* is expected to be equal to:
		((method->invoker == sym_invokeJITCompiledMethod) ||
		 (method->invoker == sym_compileAndInvokeMethod))
	*/
#ifndef IGNORE_DISABLE
      && (compiler_enabled || (method->invoker != sym_compileAndInvokeMethod))
	/* not (invoker is compiler, but the method is not to be compiled) */
#endif	/* IGNORE_DISABLE */
      ) {
	/* compiled method */
    cur_frame->returnpc = (unsigned char *)stackpointer;

    CALL_INVOKER;
  }
  else {
	/* normal Java or native method */
    /* restack from native stack to JVM stack */
    stack_item *sp = stackpointer;
    stack_item *optop = cur_frame->optop;
    char *argsizes;

    argsizes = ((CodeInfo *)method->CompiledCodeInfo)->argsizes;

    while (*argsizes) {
      if (*argsizes == 1) {
	optop[0] = sp[0];
	optop++;  sp--;
      }
      else {
	optop[0] = sp[-1];
	optop[1] = sp[0];
	optop += 2;  sp -= 2;
      }
      argsizes++;
    }

    CALL_INVOKER;

    if (!(access & (ACC_ABSTRACT | ACC_NATIVE | ACC_MACHINE_COMPILED))) {
	/* normal Java method */
      int exec_ret;
      stack_item *optop, *old_optop;

      exec_ret = ExecuteJava(method->code, ee);
      cur_frame = ee->current_frame;	/* reload */
      old_optop = cur_frame->optop;
#ifdef EXECUTEJAVA_IN_ASM
      if (cur_frame->monitor)  monitorExit(obj_monitor(obj));
#endif
      cur_frame = cur_frame->prev;
      ee->current_frame = cur_frame;
      if (!exec_ret)  return -1;

#ifdef EXECUTEJAVA_IN_ASM
      /* These operation is required
	 only with x86 assembly ver. of executeJava.c */
      if (retsize != 0) {
	optop = cur_frame->optop;
	if (retsize == 1) {
	  optop[0] = old_optop[-1];  optop++;
	}
	else {	/* retsize == 2 */
	  optop[0] = old_optop[-2];  optop[1] = old_optop[-1];  optop += 2;
	}
	cur_frame->optop = optop;
      }
#endif
    }	/* normal Java method */
  }	/* normal Java or native method */

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    stack_item *optop;
    optop = ee->current_frame->optop;
    if (info->ret_size > 0) {
      printf("  optop[-1]: 0x%x %d %g\n",
		optop[-1].i, optop[-1].i, optop[-1].f);
      if (info->ret_size > 1) {
	printf("  optop[-2]: 0x%x %d %g\n",
		optop[-2].i, optop[-2].i, optop[-2].f);
      }
      fflush(stdout);
    }
  }
#endif

#if 0
  return ((CodeInfo *)method->CompiledCodeInfo)->ret_size;
#else
  return retsize;
#endif
}


/*
 * returns: method to be invoked, NULL if an exception occurred.
 */
struct methodblock *getInterfaceMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee,
	JHandle *obj, struct methodblock *imethod, unsigned char *guessptr) {
  struct methodblock *method;
  ClassClass *intfClazz;
  unsigned int offset;

  ClassClass *cb;
  struct methodtable *mtable;
  struct imethodtable *imtable;
  int guess;

  intfClazz = fieldclass(&imethod->fb);
  offset = imethod->fb.u.offset;
  guess = *guessptr;
#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("  intf method: %s#%s %s, guess: %d\n",
	cbName(intfClazz), imethod->fb.name, imethod->fb.signature,
	guess);
    fflush(stdout);
  }
#endif

  if (!obj) {
    SignalError(ee, JAVAPKG "NullPointerException", 0);
    return NULL;
  }

  if (obj_flags(obj) == T_NORMAL_OBJECT) {
    mtable = obj_methodtable(obj);  cb = mtable->classdescriptor;
#ifdef JITDO
    if ((mtable == proxy_methodtable) && (REMOTE_FLAG(ee))) {
      /* obj instanceof Proxy*/
      cb = unhand((HNET_shudo_jitdo_Proxy *)obj)->clazz;
      mtable = cbMethodTable(cb);
    }
#endif
  }
  else {
    cb = classJavaLangObject;  mtable = cbMethodTable(cb);
  }
  imtable = cbIntfMethodTable(cb);

  /* `guess' is already calculated. */
  if (guess >= 0 && guess < imtable->icount &&
	imtable->itable[guess].classdescriptor == intfClazz) {
    goto getintf_done;
  }

  if (guess < 0 || guess >= imtable->icount ||
	imtable->itable[guess].classdescriptor != intfClazz) {
    for (guess = imtable->icount - 1; ;guess--) {
      if (guess < 0) {
	invokeInterfaceError(ee, NULL, cb, intfClazz);
	return NULL;
      }
      if (imtable->itable[guess].classdescriptor == intfClazz) {
	*guessptr = (unsigned char)guess;
	break;
      }
    }
  }

getintf_done:
  method = mt_slot(mtable, imtable->itable[guess].offsets[offset]);

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("  getInterfaceMethod() returns %s#%s\n",
		cbName(method->fb.clazz), method->fb.name);
    fflush(stdout);
  }
#endif

  return method;
}


JHandle *multianewarray(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee, int dimensions, ClassClass *arrayclazz,
	stack_item *stackpointer) {
  stack_item *optop = ee->current_frame->optop;
  int i;

  stackpointer += (dimensions - 1);
  for (i = 0; i < dimensions; i++) {
    int size = stackpointer->i;
#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      printf("  %d: %d\n", i, size);
      fflush(stdout);
    }
#endif
#ifndef NO_CHECK
    if (size < 0)	/* NegativeArraySizeException */
      return (JHandle *)-1;
#endif
    optop[i].i = size;
    stackpointer--;
  }

  return MultiArrayAlloc(dimensions, arrayclazz, optop);
}



struct CatchFrame *searchCatchFrame(
	ExecEnv *ee, struct methodblock *mb, int bytepcoff
#ifdef RUNTIME_DEBUG
	, int runtime_debug
#endif
) {
  ClassClass *methodClazz;
  cp_item_type *constant_pool;
  unsigned char *type_table;
  struct CatchFrame *cf;
  int found;
  int i;

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("  searchCatchFrame() called:\n"
	   "    ee: 0x%x, mb: 0x%x, 0x%x\n",
	(int)ee, (int)mb, bytepcoff);
    fflush(stdout);
  }
#endif

  methodClazz = fieldclass(&mb->fb);
  constant_pool = cbConstantPool(methodClazz);
  type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    JHandle *exc = ee->exception.exc;
    printf("  exc: ");
    if (exc)
      printf("0x%08x(%s)\n",
	(int)exc, cbName(exc->methods->classdescriptor));
    else
      printf(" is NULL !!!\n");
    printf("  bytepcoff %d\n", bytepcoff);
    fflush(stdout);
  }
#endif

  found = 0;	/* false */
  cf = mb->exception_table;
  for (i = 0; i < mb->exception_table_length; i++) {
#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      printf("  pc: start %ld end %ld type %d\n",
		cf->start_pc, cf->end_pc, cf->catchType);
      fflush(stdout);
    }
#endif
    if ((cf->start_pc <= bytepcoff) && (cf->end_pc > bytepcoff)) {
      int typeindex = cf->catchType;
      char *catchName;
      ClassClass *catchClazz = NULL;
      ClassClass *excClazz;

      if (typeindex == 0) {
		/* This catch frame accpets any type of exceptions. */
	found = 1;  goto search_finish;
      }

      catchName = GetClassConstantClassName(constant_pool, typeindex);
#ifdef RUNTIME_DEBUG
      if (runtime_debug) {
	printf("    catch name: %s\n", catchName);
	fflush(stdout);
      }
#endif

      for (excClazz = obj_array_classblock(ee->exception.exc);
		excClazz; excClazz = cbSuperclass(excClazz)) {
	if ( !(strcmp(cbName(excClazz), catchName)) &&
		(cbLoader(excClazz) == cbLoader(methodClazz)) ) {
	  found = 1;  goto search_finish;
	}

	if (!catchClazz) {
	  if (!(type_table[typeindex] & CONSTANT_POOL_ENTRY_RESOLVED))
	    if (!(ResolveClassConstantFromClass(methodClazz, typeindex, ee,
			1 << CONSTANT_Class))) {
#ifdef RUNTIME_DEBUG
	      if (runtime_debug) {
		printf("    resolution of catch class failed: %d\n",
			typeindex);
		fflush(stdout);
	      }
#endif
	      continue;
	    }
	  catchClazz = constant_pool[typeindex].clazz;
	}
	if (excClazz == catchClazz) {
	  found = 1;  goto search_finish;
	}
      }
    }	/* if (start_pc <= pc < end_pc) */

    cf++;
  }	/* for (i = 0; ... */

search_finish:
  if (found) {
#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      printf("  exc. handler is found: %s#%s %ld-%ld type:%d\n"
		"  native offset 0x%x(%d).\n",
		cbName(mb->fb.clazz), mb->fb.name,
		cf->start_pc, cf->end_pc, (int)cf->catchType,
		(int)cf->compiled_CatchFrame, (int)cf->compiled_CatchFrame);
      fflush(stdout);
    }
#endif
    exceptionClear(ee);

    /* push exception object */
    ee->current_frame->optop = ee->current_frame->ostack;
    /* ee->current_frame->optop = ... */
    (ee->current_frame->optop++)->h = ee->exception.exc;

    return cf;
  }
  else {
#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      printf("  exc. handler is not found at method %s#%s.\n",
		cbName(mb->fb.clazz), mb->fb.name);
      fflush(stdout);
    }
#endif
    return NULL;
  }
}


void showStackFrames(ExecEnv *ee) {
  JavaFrame *frame = ee->current_frame;
  printf("stack frames:\n");
  while (frame) {
    struct methodblock *frameMb = frame->current_method;
    if (frameMb) {
      printf("  %s#%s %s 0x%x  ",
	cbName(frameMb->fb.clazz), frameMb->fb.name, frameMb->fb.signature,
	(int)frameMb);
      if (frame->lastpc)
	printf("pc: %d", frame->lastpc - frameMb->code);
      if (frameMb->fb.access & ACC_MACHINE_COMPILED)
	printf(" (compiled)");
      printf("\n");
    }
    else {
      printf("  (null)  lastpc: 0x%08x\n", (int)frame->lastpc);
    }

    fflush(stdout);

    frame = frame->prev;
  }
}


#ifdef RUNTIME_DEBUG
typedef struct Classjava_lang_Throwable {
  struct Hjava_lang_Object *backtrace;
  struct Hjava_lang_String *detailMessage;
} Classjava_lang_Throwable;
HandleTo(java_lang_Throwable);

static struct methodblock *pc2method(unsigned char *pc);

void showExcStackTrace(JHandle *o) {
  HArrayOfByte *backtrace =
	(HArrayOfByte *)(unhand((Hjava_lang_Throwable *)o)->backtrace);
  if (backtrace) {
    unsigned char **data = (unsigned char **)(unhand(backtrace)->body);
    unsigned char **end = (unsigned char **)(data + obj_length(backtrace));
    for (; data < end; data++) {
      if (*data) {
#define BUF_SIZE	256
	char buf[BUF_SIZE];
	char *p = buf;
	struct methodblock *mb = pc2method(*data);

	strncpy(buf, "\tat ", 4);  p += 4;
	if (mb) {
	  p += snprintf(p, buf + BUF_SIZE - p, "%s#%s %s  ",
		cbName(fieldclass(&mb->fb)), mb->fb.name, mb->fb.signature);
	  p += snprintf(p, buf + BUF_SIZE - p, "pc: %d",
		*data - mb->code);
	  if (mb->fb.access & ACC_MACHINE_COMPILED)
	    p += snprintf(p, buf + BUF_SIZE - p, "(compiled)");
	}
	else {
	  p += snprintf(p, buf + BUF_SIZE - p, "(null)");
	}
	printf(buf);
	printf("\n");
      }
    }
  }
}


static struct methodblock *pc2method(unsigned char *pc) {
  struct methodblock *mb;
  ClassClass *cb;
  int i, j;

  BINCLASS_LOCK();
  for (i = nbinclasses; --i >= 0; ) {
    cb = binclasses[i];
    for (mb = cbMethods(cb), j = cbMethodsCount(cb); --j >= 0; mb++) {
      if (!(mb->fb.access & ACC_NATIVE)) {		/* not native method */
	if (mb->fb.access & ACC_MACHINE_COMPILED) {	/* compiled method */
	  if (PCinCompiledCode(pc, mb))
	    goto done;
	} else { 
	  unsigned char *code = (unsigned char *)mb->code;
	  if (pc >= code && pc < code + mb->code_length)
	    goto done;
	}
      }
    }
  }
  mb = 0;

done:
  BINCLASS_UNLOCK();
  return mb;
}
#endif	/* RUNTIME_DEBUG */


#ifdef RUNTIME_DEBUG
char *showObjectBody(char *sig, JHandle *obj) {
  switch (*sig) {
  case '[':
    sig++;
    if (*sig == 'L')  while (*(sig++) != ';');
    else sig++;
    if (!obj)
      printf("(null)");
    else {
      char buf[256];
show_array:
      printf("len %ld", obj_length(obj));
      switch (obj_flags(obj)) {
      case T_BYTE:
	printf(" `%s'", unhand((HArrayOfByte *)obj)->body);
	break;
      case T_CHAR:
	printf(" `%s'",
		unicode2utf(unhand((HArrayOfChar *)obj)->body, obj_length(obj),
			 buf, 256));
	break;
      }
    }
    break;
  case 'L':
    while (*(sig++) != ';');
    if (!obj)
      printf("(null)");
    else {
      char *classname;
      if (obj_flags(obj) != T_NORMAL_OBJECT)  goto show_array;
      classname = cbName(obj_array_classblock(obj));
      printf("class %s", classname);
      if (!strcmp(classname, "java/lang/String")) {
	int strlen = javaStringLength((Hjava_lang_String *)obj);
	printf(" len %d `", strlen);
	if (strlen)  javaStringPrint((Hjava_lang_String *)obj);
	printf("'");
      }
    }
    break;
  default:
    sig++;
  }

  return sig;
}
#endif	/* RUNTIME_DEBUG */
